/*
 * Copyright 2003, Franz-Josef Elmer, All rights reserved
 *
 * This library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details
 * (http://www.gnu.org/copyleft/lesser.html).
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package jcckit;

import java.awt.BorderLayout;
import java.awt.Canvas;
import java.awt.Dimension;
import java.awt.Frame;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;

import jcckit.data.DataPlot;
import jcckit.graphic.GraphPoint;
import jcckit.graphic.GraphicalElement;
import jcckit.graphic.Renderer;
import jcckit.plot.Plot;
import jcckit.plot.PlotCanvas;
import jcckit.plot.PlotEvent;
import jcckit.renderer.GraphicsRenderer;
import jcckit.renderer.Transformation;
import jcckit.util.ConfigParameters;
import jcckit.util.PropertiesBasedConfigData;

/**
 * Class which handles plotting into a <tt>Graphics</tt> context based on
 * the {@link jcckit.renderer.GraphicsRenderer}.
 * This class is not a subclass of <tt>java.awt.Component</tt>.
 * The actual AWT component presenting the plot is an innerclass. Its
 * instance wrapped by <tt>GraphicsPlotCanvas</tt> can be obtained with
 * {@link #getGraphicsCanvas}.
 * <p>
 * The plot is painted by using double-buffering and pre-rendered
 * view of the coordinate system. That is, the coordinate system is
 * drawn into an off-screen image. It will be redrawn only if the 
 * size of the embedding AWT component is changed.
 * 
 * @author Franz-Josef Elmer
 */
public class GraphicsPlotCanvas extends  PlotCanvas {
  /** Key of a configuration parameter. */
  public static final String BACKGROUND_KEY = "background",
                             FOREGROUND_KEY = "foreground";
  /** 
   * AWT component which actually shows the rendered plot. 
   * 
   * @author Franz-Josef Elmer
   */
  protected class GraphicsCanvas extends Canvas {
    private Image _buffer;
    private Image _coordinateSystem;
    private Graphics _bufferGraphics;
    private Dimension _currentSize;

    /** Handles the event. Just invokes <tt>repaint()</tt>. */
    public void handleEvent(PlotEvent event) {
      repaint();
    }

    /** Directly invokes <tt>paint</tt>. */
    public void update(Graphics g) {
      paint(g);
    }

    /** 
     * Paints the plot by using double-buffering and pre-rendered view of
     * the coordinate system.
     */
    public void paint(Graphics g) {
      Dimension size = getSize();
      if (!size.equals(_currentSize)) {
        _currentSize = size;
        if (_buffer != null) {
          _buffer.flush();
          _bufferGraphics.dispose();
        }
        _buffer = createImage(size.width, size.height);
        _bufferGraphics = _buffer.getGraphics();
        if (_coordinateSystem != null) {
          _coordinateSystem.flush();
          _coordinateSystem = null;
        }
        calculateTransformation(size);
      }
      //long time = System.currentTimeMillis();
      Plot plot = getPlot();
      if (_coordinateSystem == null) {
        _coordinateSystem = createImage(size.width, size.height);
        Graphics csg = _coordinateSystem.getGraphics();
        csg.setColor(getBackground());
        csg.fillRect(0, 0, size.width, size.height);
        csg.setColor(getForeground());
        plot.getCoordinateSystem().renderWith(createRenderer(csg));
        csg.dispose();
      }
      prepare(_bufferGraphics);
      _bufferGraphics.drawImage(_coordinateSystem, 0, 0, null);
      Renderer renderer = createRenderer(_bufferGraphics);
      GraphicalElement[] curves = plot.getCurves();
      for (int i = 0; i < curves.length; i++) {
        curves[i].renderWith(renderer);
      }
      if (plot.isLegendVisible()) {
        plot.getLegend().renderWith(renderer);
      }
      g.drawImage(_buffer, 0, 0, null);
      //System.out.println((System.currentTimeMillis() - time) + " msec");
    }
    
    /** 
     * Prepare graphics context before drawing the pre-rendered view of
     * the coordinate system. Does nothing but will be used in subclasses.
     */
    protected void prepare(Graphics g) {}

    /** 
     * Calculate the transformation form device-independent coordinates
     * into device-dependent coordinates according to the specified
     * canvas size.
     */
    protected void calculateTransformation(Dimension size) {
      _transformation = new Transformation(size.width, size.height,
              getPaper(), getHorizontalAnchor(), getVerticalAnchor());
    }
    
    /** 
     * Creates an appropriated {@link Renderer} for the specified
     * <tt>Graphics</tt> context.
     */
    protected Renderer createRenderer(Graphics g) {
      return new GraphicsRenderer(g, this, _transformation);
    }
    
  } // end class GraphicsCanvas

  private Transformation _transformation;

  /** Wrapped AWT component. */
  protected GraphicsCanvas _canvas;

  /**
   * Creates an instance from the specfied configuration parameters.
   * <table border=1 cellpadding=5>
   * <tr><th>Key &amp; Default Value</th><th>Type</th><th>Mandatory</th>
   *     <th>Description</th></tr>
   * <tr><td><tt>background = </tt><i>default background color of the wrapped
   *      AWT component</i></td>
   *     <td><tt>Color</tt></td><td>no</td>
   *     <td>Background color of the wrapped AWT component.</td></tr>
   * <tr><td><tt>foreground = </tt><i>default foreground color of the wrapped
   *      AWT component</i></td>
   *     <td><tt>Color</tt></td><td>no</td>
   *     <td>Foreground color of the wrapped AWT component.</td></tr>
   * </table>
   * In addition the configuration parameters of the
   * <a href="plot/PlotCanvas.html#PlotCanvas(jcckit.util.ConfigParameters)">
   * constructor</a> of the superclass {@link jcckit.plot.PlotCanvas} apply.
   */
  public GraphicsPlotCanvas(ConfigParameters config) {
    super(config);
    createGraphicsCanvas();
    _canvas.setBackground(config.getColor(BACKGROUND_KEY,
                                          _canvas.getBackground()));
    _canvas.setForeground(config.getColor(FOREGROUND_KEY,
                                          _canvas.getForeground()));
  }
  
  /** Creates an instance of {@link GraphicsCanvas}. */
  protected void createGraphicsCanvas() {
    _canvas = new GraphicsCanvas();
  }

  /** Repaints the wrapped AWT component. */
  public void plotChanged(PlotEvent event) {
    _canvas.handleEvent(event);
  }

  /** Returns the wrapped AWT component which contains the rendered plot. */
  public Canvas getGraphicsCanvas() {
    return _canvas;
  }
  
  /**
   * Maps the cursor position onto a point in device-independent coordinates.
   * @param x X-coordinate of the cursor.
   * @param y Y-coordinate of the cursor.
   */
  public GraphPoint mapCursorPosition(int x, int y) {
    return _transformation.transformBack(x, y);
  }

  /**
   * Shows a plot in a <tt>Frame</tt>. The plot (data and layout) is 
   * defined in the <tt>.properties</tt> file specified by the first
   * command line argument.
   * <p>
   * Usage: <tt>java jcckit.GraphicsPlotCanvas
   *            &lt;<i>properties file</i>&gt;</tt>
   */
  public static void main(String[] args) throws Exception {
    if (args.length == 0) {
      System.out.println(
          "Usage: java jcckit.GraphicsPlotCanvas <properties file>");
      System.exit(0);
    }
    ConfigParameters config
        = new ConfigParameters(new PropertiesBasedConfigData(args[0]));
    connectDataAndShow(args[0], config, new GraphicsPlotCanvas(config));
  }

  /**
   * Connects <tt>plotCanvas</tt> with data created from <tt>config</tt>
   * and shows the chart in a <tt>Frame</tt> where <tt>fileName</tt> appears
   * in the title.
   */
  protected static void connectDataAndShow(String fileName,
                                           ConfigParameters config,
                                           GraphicsPlotCanvas plotCanvas) {
    plotCanvas.connect(DataPlot.create(config));
    
    Frame frame = new Frame("JCCKit: " + fileName);
    frame.addWindowListener(new WindowAdapter() {
              public void windowClosing(WindowEvent event) {
                System.exit(0);
              }
            }
          );
    frame.add(plotCanvas.getGraphicsCanvas(), BorderLayout.CENTER);
    frame.setSize(666, 555);
    frame.show();
  }
}
